from sensiml.widgets.auto_sense import AutoSenseWidget
from sensiml.widgets.download_widget import DownloadWidget
from sensiml.widgets.query_widget import QueryWidget
from sensiml.widgets.flash_widget import FlashWidget
from sensiml.widgets.model_visualization_widget import ModelVisualizationWidget
from sensiml.widgets.model_explore_widget import ModelExploreWidget
from sensiml.widgets.dashboard_widget import DashBoard


__all__ = ['AutoSenseWidget', 'DownloadWidget', 'QueryWidget', 'FlashWidget',
           "ModelVisualizationWidget", "DashBoard", "ModelExploreWidget"]
